/*
 * Buxton
 *
 * Copyright (C) 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License)
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <pwd.h>
#include <sys/types.h>
#include <unistd.h>
#include <inttypes.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "common.h"
#include "direct.h"

#include "c_log.h"
#include "c_common.h"
#include "c_direct.h"

#define BXT_LOG_FOLDER "/run/buxton2/log"
#define BXT_NULL_STRING "(VCONF_NULL)"

static FILE *_fp;
static int _log_max_line = 2000;
static bool _log_on = true;

static const char *confpath;
static const char const *type_names[BUXTON_TYPE_MAX] = {
	[BUXTON_TYPE_UNKNOWN]   = "Unknown",
	[BUXTON_TYPE_STRING]    = "String",
	[BUXTON_TYPE_INT32]     = "Int32",
	[BUXTON_TYPE_UINT32]    = "UInt32",
	[BUXTON_TYPE_INT64]     = "Int64",
	[BUXTON_TYPE_UINT64]    = "UInt64",
	[BUXTON_TYPE_DOUBLE]    = "Double",
	[BUXTON_TYPE_BOOLEAN]   = "Boolean",
};

void c_direct_set_conf(const char *conf)
{
	if (!conf || !*conf) {
		bxt_err("Invalid config path. Default path is used.");
		return;
	}

	confpath = conf;
}

static void c_exit(void)
{
	direct_exit();
}

static int c_init(void)
{
	int r;

	r = direct_init(MODULE_DIR, confpath ? confpath : CONFPATH, &_log_on, &_log_max_line);
	if (r != BUXTON_ERROR_NONE)
		bxt_err("failed to init %s", confpath);

	return r;
}

int c_direct_get(const struct buxton_layer *layer,
		const char *key, UNUSED const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	int r;
	struct buxton_value val = { BUXTON_TYPE_UNKNOWN, };

	if (!layer || !key || !*key) {
		bxt_err("Get: Layer '%s' Key '%s': Invalid parameter",
				layer ? buxton_layer_get_name(layer) : "",
				key ? key : "");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = c_init();
	if (r != BUXTON_ERROR_NONE)
		return r;

	r = direct_get(layer, key, &val);
	c_exit();
	if (r != BUXTON_ERROR_NONE) {
		bxt_err("Get: Layer '%s' Key '%s': %s %d",
			buxton_layer_get_name(layer), key, buxton_err_get_str(r), r);
		return r;
	}
	c_print_value(layer, key, &val);
	value_free(&val);

	return BUXTON_ERROR_NONE;
}

static bool _init_log()
{
	char buffer[256] = {0, };
	int ret;

	ret = mkdir(BXT_LOG_FOLDER, 0755);
	if (ret < 0  && errno != EEXIST) {
		bxt_err("Failed to mkdir %s - %d", BXT_LOG_FOLDER, errno);
		return false;
	}

	snprintf(buffer, sizeof(buffer),
			"%s/direct_log_%d.log", BXT_LOG_FOLDER, getpid());
	_fp = fopen(buffer, "wt");
	if (_fp == NULL) {
		bxt_err("Failed to open %s - %d", buffer, errno);
		return false;
	}

	return true;
}

static void _write_file_log(const char *key, const char *value)
{
	char buf[512] = {0,};
	char time_buf[32] = {0,};
	int ret = 0;
	int fd;
	time_t now;
	static int _log_index = 0;
	char app_id[512] = {0,};
	const char *refined_value;

	if (_log_on == false)
		return;

	if (_fp == NULL) {
		if (_init_log() == false)
			return;
	}

	snprintf(buf, sizeof(buf), "/proc/%d/cmdline", getppid());

	fd = open(buf, O_RDONLY);
	if (fd >= 0) {
		ret = read(fd, app_id, sizeof(app_id) - 1);
		if (ret < 0)
			bxt_err("Cannot read %s: %d", buf, errno);
		close(fd);
	}

	time(&now);
	ctime_r(&now, time_buf);

	/* remove newline */
	if (strlen(time_buf) > 1)
		time_buf[strlen(time_buf) -1] = '\0';

	if (value == NULL || strlen(value) == 0)
		refined_value = BXT_NULL_STRING;
	else
		refined_value = value;

	_log_index++;
	snprintf(buf, sizeof(buf), "\n[%5d]%s DirectSet - ppid(%d:%s) - %s ",
		_log_index, time_buf, getppid(), app_id, key);
	ret = fwrite(buf, strlen(buf), 1 , _fp);
	if (ret < 0) {
		bxt_err("Cannot write the key log: %d, %s %s", errno, buf, refined_value);
		return;
	}

	ret = fwrite(refined_value, strlen(refined_value), 1 , _fp);
	if (ret < 0) {
		bxt_err("Cannot write the value log: %d, %s %s", errno, buf, refined_value);
		return;
	}
}

static int c_direct_set(const struct buxton_layer *layer,
		const char *key, const char *value, enum buxton_key_type type)
{
	int r;
	struct buxton_value val;

	if (!layer || !key || !*key || !value) {
		bxt_err("Set: Layer '%s' Key '%s' Value '%s': Invalid parameter",
				layer ? buxton_layer_get_name(layer) : "",
				key ? key : "", value ? value : "");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = c_set_value(type, value, &val);
	if (r != BUXTON_ERROR_NONE)
		return r;

	r = c_init();
	if (r != BUXTON_ERROR_NONE)
		return r;

	r = direct_set(layer, key, &val);
	c_exit();
	if (r != BUXTON_ERROR_NONE) {
		bxt_err("Set: Layer '%s' Key '%s' Value '%s': %s %d",
		buxton_layer_get_name(layer), key, value, buxton_err_get_str(r), r);
		return r;
	}

	_write_file_log(key, value);

	return r;
}

int c_direct_set_str(const struct buxton_layer *layer,
		const char *key, const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	return c_direct_set(layer, key, value, BUXTON_TYPE_STRING);
}

int c_direct_set_int32(const struct buxton_layer *layer,
		const char *key, const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	return c_direct_set(layer, key, value, BUXTON_TYPE_INT32);
}

int c_direct_set_uint32(const struct buxton_layer *layer,
		const char *key, const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	return c_direct_set(layer, key, value, BUXTON_TYPE_UINT32);
}

int c_direct_set_int64(const struct buxton_layer *layer,
		const char *key, const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	return c_direct_set(layer, key, value, BUXTON_TYPE_INT64);
}

int c_direct_set_uint64(const struct buxton_layer *layer,
		const char *key, const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	return c_direct_set(layer, key, value, BUXTON_TYPE_UINT64);
}

int c_direct_set_double(const struct buxton_layer *layer,
		const char *key, const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	return c_direct_set(layer, key, value, BUXTON_TYPE_DOUBLE);
}

int c_direct_set_bool(const struct buxton_layer *layer,
		const char *key, const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	return c_direct_set(layer, key, value, BUXTON_TYPE_BOOLEAN);
}


static int c_direct_create(const struct buxton_layer *layer,
		const char *key, const char *value, enum buxton_key_type type,
		const char *rpriv, const char *wpriv)
{
	int r;
	struct buxton_value val;

	if (!layer || !key || !*key || !value || !rpriv || !wpriv) {
		bxt_err("Create: '%s' '%s' '%s' Priv '%s' '%s': Invalid parameter",
				layer ? buxton_layer_get_name(layer) : "",
				key ? key : "", value ? value : "",
				rpriv ? rpriv : "", wpriv ? wpriv : "");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = c_set_value(type, value, &val);
	if (r != BUXTON_ERROR_NONE)
		return r;

	r = c_init();
	if (r != BUXTON_ERROR_NONE)
		return r;
	bxt_err("Create: ");

	r = direct_create(layer, key, rpriv, wpriv, &val);

	c_exit();

	if (r != BUXTON_ERROR_NONE)
		bxt_err("Create: '%s' '%s' '%s' Priv '%s' '%s': %s %d",
				buxton_layer_get_name(layer), key, value,
				rpriv, wpriv, buxton_err_get_str(r), r);

	return r;
}

int c_direct_create_str(const struct buxton_layer *layer,
		const char *key, const char *value,
		const char *rpriv, const char *wpriv)
{
	return c_direct_create(layer, key, value, BUXTON_TYPE_STRING,
			rpriv, wpriv);
}

int c_direct_create_int32(const struct buxton_layer *layer,
		const char *key, const char *value,
		const char *rpriv, const char *wpriv)
{
	return c_direct_create(layer, key, value, BUXTON_TYPE_INT32,
			rpriv, wpriv);
}

int c_direct_create_uint32(const struct buxton_layer *layer,
		const char *key, const char *value,
		const char *rpriv, const char *wpriv)
{
	return c_direct_create(layer, key, value, BUXTON_TYPE_UINT32,
			rpriv, wpriv);
}

int c_direct_create_int64(const struct buxton_layer *layer,
		const char *key, const char *value,
		const char *rpriv, const char *wpriv)
{
	return c_direct_create(layer, key, value, BUXTON_TYPE_INT64,
			rpriv, wpriv);
}

int c_direct_create_uint64(const struct buxton_layer *layer,
		const char *key, const char *value,
		const char *rpriv, const char *wpriv)
{
	return c_direct_create(layer, key, value, BUXTON_TYPE_UINT64,
			rpriv, wpriv);
}

int c_direct_create_double(const struct buxton_layer *layer,
		const char *key, const char *value,
		const char *rpriv, const char *wpriv)
{
	return c_direct_create(layer, key, value, BUXTON_TYPE_DOUBLE,
			rpriv, wpriv);
}

int c_direct_create_bool(const struct buxton_layer *layer,
		const char *key, const char *value,
		const char *rpriv, const char *wpriv)
{
	return c_direct_create(layer, key, value, BUXTON_TYPE_BOOLEAN,
			rpriv, wpriv);
}

static int c_direct_get_priv(const struct buxton_layer *layer,
		const char *key, enum buxton_priv_type type)
{
	int r;
	char *priv;

	if (!layer || !key || !*key) {
		bxt_err("Get-priv: Layer '%s' Key '%s': Invalid parameter",
				layer ? buxton_layer_get_name(layer) : "",
				key ? key : "");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = c_init();
	if (r != BUXTON_ERROR_NONE)
		return r;

	r = direct_get_priv(layer, key, type, &priv);
	c_exit();
	if (r != BUXTON_ERROR_NONE) {
		bxt_err("Get-priv: Layer '%s' Key '%s': %s %d",
				buxton_layer_get_name(layer), key, buxton_err_get_str(r), r);
		return r;
	}

	c_print_priv(layer, key, type, priv);
	free(priv);

	return r;
}

int c_direct_get_rpriv(const struct buxton_layer *layer,
		const char *key, UNUSED const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	return c_direct_get_priv(layer, key, BUXTON_PRIV_READ);
}

int c_direct_get_wpriv(const struct buxton_layer *layer,
		const char *key, UNUSED const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	return c_direct_get_priv(layer, key, BUXTON_PRIV_WRITE);
}

static int c_direct_set_priv(const struct buxton_layer *layer,
		const char *key, const char *priv, enum buxton_priv_type type)
{
	int r;

	if (!layer || !key || !*key || !priv) {
		bxt_err("Set-priv: Layer '%s' Key '%s' Priv. '%s': Invalid parameter",
				layer ? buxton_layer_get_name(layer) : "",
				key ? key : "", priv ? priv : "");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = c_init();
	if (r != BUXTON_ERROR_NONE)
		return r;

	r = direct_set_priv(layer, key, type, priv);
	c_exit();
	if (r != BUXTON_ERROR_NONE) {
		bxt_err("Set-priv: Layer '%s' Key '%s' Priv. '%s': %s %d",
				buxton_layer_get_name(layer), key, priv, buxton_err_get_str(r), r);
	}

	return r;
}

int c_direct_set_rpriv(const struct buxton_layer *layer,
		const char *key, const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	return c_direct_set_priv(layer, key, value, BUXTON_PRIV_READ);
}

int c_direct_set_wpriv(const struct buxton_layer *layer,
		const char *key, const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	return c_direct_set_priv(layer, key, value, BUXTON_PRIV_WRITE);
}

int c_direct_unset(const struct buxton_layer *layer,
		const char *key, UNUSED const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	int r;

	if (!layer || !key || !*key) {
		bxt_err("Unset: Layer '%s' Key '%s': Invalid parameter",
				layer ? buxton_layer_get_name(layer) : "",
				key ? key : "");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = c_init();
	if (r != BUXTON_ERROR_NONE)
		return r;

	r = direct_unset(layer, key);
	c_exit();
	if (r != BUXTON_ERROR_NONE) {
		bxt_err("Unset: Layer '%s' Key '%s': %s %d",
				buxton_layer_get_name(layer), key, buxton_err_get_str(r), r);
	}

	return r;
}

int c_direct_list(const struct buxton_layer *layer,
		UNUSED const char *key, UNUSED const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	int r;
	char **keys;
	char **k;

	if (!layer) {
		bxt_err("List: Layer '%s': Invalid parameter",
				layer ? buxton_layer_get_name(layer) : "");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = c_init();
	if (r != BUXTON_ERROR_NONE)
		return r;

	r = direct_list(layer, &keys, NULL, true);
	c_exit();
	if (r != BUXTON_ERROR_NONE) {
		bxt_err("List: Layer '%s': %s %d", buxton_layer_get_name(layer),
				 buxton_err_get_str(r), r);
		return r;
	}

	k = keys;
	while (k && *k) {
		printf("%s\n", *k);
		k++;
	}

	buxton_free_keys(keys);

	return r;
}

int c_direct_remove_garbage(const struct buxton_layer *layer,
		UNUSED const char *key, UNUSED const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	struct buxton_layer *normal_layer;
	struct buxton_layer *base_layer;
	struct buxton_value val = { BUXTON_TYPE_UNKNOWN, };
	char **keys;
	unsigned int len;
	int r;
	int i;

	if (!layer) {
		bxt_err("Remove garbage data : Layer '%s' : Invalid parameter",
				layer ? buxton_layer_get_name(layer) : "");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = c_init();
	if (r != BUXTON_ERROR_NONE)
		return r;

	normal_layer = buxton_create_layer(layer->name);
	buxton_layer_set_type(normal_layer, BUXTON_LAYER_NORMAL);
	base_layer = buxton_create_layer(layer->name);
	buxton_layer_set_type(base_layer, BUXTON_LAYER_BASE);

	r = direct_list(normal_layer, &keys, &len, false);
	if (r != BUXTON_ERROR_NONE) {
		buxton_free_layer(normal_layer);
		buxton_free_layer(base_layer);
		bxt_err("Remove garbage data: Layer '%s' : %s %d",
						buxton_layer_get_name(normal_layer),
						buxton_err_get_str(r), r);
		return r;
	}

	i = 0;
	while (i < len) {
		r = direct_get(base_layer, keys[i], &val);
		if (r == BUXTON_ERROR_NOT_EXIST) {
			r = direct_unset(normal_layer, keys[i]);
			if (r != BUXTON_ERROR_NONE) {
				bxt_err("Remove garbage data: Layer '%s' Key '%s': %s %d",
						buxton_layer_get_name(normal_layer), keys[i],
						buxton_err_get_str(r), r);
			}
		}
		value_free(&val);
		i++;
	}

	c_exit();
	buxton_free_layer(normal_layer);
	buxton_free_layer(base_layer);

	return BUXTON_ERROR_NONE;
}

int c_direct_dump(const struct buxton_layer *layer,
		UNUSED const char *key, UNUSED const char *value,
		UNUSED const char *rpriv, UNUSED const char *wpriv)
{
	int r, key_len, i;
	struct buxton_value *value_list;
	char **key_list;

	if (!layer) {
		bxt_err("List: Layer '%s': Invalid parameter",
				layer ? buxton_layer_get_name(layer) : "");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	r = c_init();
	if (r != BUXTON_ERROR_NONE)
		return r;

	r = direct_dump(layer, &key_list, &value_list, &key_len);
	c_exit();
	if (r != BUXTON_ERROR_NONE) {
		bxt_err("List: Layer '%s': %s %d", buxton_layer_get_name(layer),
				buxton_err_get_str(r), r);
		return r;
	}

	for (i = 0 ; i < key_len; i++) {
		printf("%s, value = ", key_list[i]);

		switch (value_list[i].type) {
		case BUXTON_TYPE_STRING:
			printf("%s (%s)\n", value_list[i].value.s, type_names[value_list[i].type]);
			free(value_list[i].value.s);
			break;
		case BUXTON_TYPE_INT32:
			printf("%d (%s)\n", value_list[i].value.i, type_names[value_list[i].type]);
			break;
		case BUXTON_TYPE_UINT32:
			printf("%u (%s)\n", value_list[i].value.u, type_names[value_list[i].type]);
			break;
		case BUXTON_TYPE_INT64:
			printf("%" PRId64 " (%s)\n", value_list[i].value.i64, type_names[value_list[i].type]);
			break;
		case BUXTON_TYPE_UINT64:
			printf("%" PRIu64 " (%s)\n", value_list[i].value.u64, type_names[value_list[i].type]);
			break;
		case BUXTON_TYPE_DOUBLE:
			printf("%lf (%s)\n", value_list[i].value.d, type_names[value_list[i].type]);
			break;
		case BUXTON_TYPE_BOOLEAN:
			printf("%s (%s)\n", value_list[i].value.b ? "True" : "False", type_names[value_list[i].type]);
			break;
		default:
			printf("Unknown type\n");
			break;
		}
	}

	buxton_free_keys(key_list);
	free(value_list);

	return BUXTON_ERROR_NONE;
}

