/*
 * Buxton
 *
 * Copyright (C) 2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>

#include "common.h"
#include "log.h"


int layer_create(const char *layer_name, struct buxton_layer **layer)
{
	struct buxton_layer *tmp_layer;

	if (!layer_name || !*layer_name || !layer) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	tmp_layer = calloc(1, sizeof(*tmp_layer));
	if (!tmp_layer) {
		bxt_err("out of memory");
		return BUXTON_ERROR_OUT_OF_MEMORY;
	}

	tmp_layer->name = strdup(layer_name);
	if (!tmp_layer->name) {
		free(tmp_layer);
		bxt_err("out of memory");
		return BUXTON_ERROR_OUT_OF_MEMORY;
	}

	tmp_layer->refcnt = 1;
	tmp_layer->uid = getuid();
	tmp_layer->type = BUXTON_LAYER_NORMAL;

	*layer = tmp_layer;

	return BUXTON_ERROR_NONE;
}

/* ignore ref. count */
void layer_free(struct buxton_layer *layer)
{
	if (!layer)
		return;

	free(layer->name);
	free(layer);
}

struct buxton_layer *layer_ref(struct buxton_layer *layer)
{
	if (!layer) {
		bxt_err("Invalid parameter");
		return NULL;
	}

	layer->refcnt++;

	return layer;
}

struct buxton_layer *layer_unref(struct buxton_layer *layer)
{
	if (!layer) {
		bxt_err("Invalid parameter");
		return NULL;
	}

	layer->refcnt--;
	if (layer->refcnt == 0) {
		layer_free(layer);
		return NULL;
	}

	return layer;
}

void value_free(struct buxton_value *val)
{
	if (!val)
		return;

	switch (val->type) {
	case BUXTON_TYPE_STRING:
		if (val->value.s) {
			free(val->value.s);
			val->value.s = NULL;
		}
		break;
	default:
		break;
	}
}

int get_search_key(const struct buxton_layer *layer, const char *key,
		const char *uid, char **layer_key)
{
	char *lykey;
	int keylen;

	if (!layer || !key || !*key || !layer_key) {
		bxt_err("Invalid parameter");
		return BUXTON_ERROR_INVALID_PARAMETER;
	}

	keylen = strlen(layer->name) + strlen(key) + 2;

	if (uid)
		keylen += strlen(uid) + 1;

	lykey = malloc(keylen);
	if (!lykey) {
		bxt_err("out of memory");
		return BUXTON_ERROR_OUT_OF_MEMORY;
	}

	snprintf(lykey, keylen, "%s\t%s%s%s", layer->name, key,
			uid ? "\t" : "", uid ? uid : "");
	*layer_key = lykey;

	return BUXTON_ERROR_NONE;
}

